from ..utils.translatable import Translatable


class Function(Translatable):
    def __init__(self, name: str, args: list[Translatable]):
        self.name: str = name
        self.args: list[Translatable] = args

    def translate(
        self,
        hyp_name: str = None,
        proof: str = None,
        **kwargs,
    ) -> str:
        if bool(hyp_name) != bool(proof):
            raise ValueError(
                "Either both or neither of hyp_name and proof must be specified."
            )

        hyp_name = f"{hyp_name} : " if hyp_name else ""
        proof = f" := {proof}" if proof else ""

        return (
            hyp_name
            + self.name
            + " ("
            + ") (".join([arg.translate() for arg in self.args])
            + ")"
            + proof
        )



import platform
import os
import shutil
from ..translator import DEFAULT_PATHS, GIT_LOCATION, LEAN_PROJECT_GIT_REPO


def update_git():
    try:
        path = DEFAULT_PATHS[platform.system()]
    except KeyError:
        raise Exception(
            f"Unsupported platform: `{platform.system()}`, please report this bug."
        )

    if path.exists():
        # run git pull
        print("Updating lean project template...\n")
        os.system(
            f'git --work-tree="{path}" --git-dir="{path/GIT_LOCATION}" reset --hard'
        )
        os.system(f'git --work-tree="{path}" --git-dir="{path/GIT_LOCATION}" pull')
        print()

    else:
        # run git clone
        path.mkdir(parents=True)
        print("Downloading lean project template...\n")
        os.system(f"git clone {LEAN_PROJECT_GIT_REPO} {path}")
        print()


def reset_git():
    try:
        path = DEFAULT_PATHS[platform.system()]
    except KeyError:
        raise Exception(
            f"Unsupported platform: `{platform.system()}`, please report this bug."
        )

    # remove the folder if it exists
    if path.exists():
        shutil.rmtree(path)

    # run git clone
    path.mkdir(parents=True)
    print("Downloading lean project template...\n")
    os.system(f"git clone {LEAN_PROJECT_GIT_REPO} {path}")
    print()



import re
from typing import Callable
from .have import Have
from .statement import Statement
from .induction import Induction
from .simple_statements import SimpleStatement
from ...utils.exceptions import MatchingError


space = r"\s*"


def _contraposition_proof(last_hyp, **kwargs) -> str:
    # if last_hyp is None:
    #     raise ValueError("hyp_name should have been provided, please report this bug.")
    return "\n".join(
        [
            f"revert {last_hyp}",
            f"rw [not_imp_not.symm]",
            f"repeat rw [not_not]",
            f"intro {last_hyp}",
        ]
    )


def _contradiction_proof(hyp_name, **kwargs) -> str:
    # if hyp_name is None:
    #     raise ValueError("hyp_name should have been provided, please report this bug.")
    return "\n".join(
        [
            f"apply not_not.mp",
            f"intro {hyp_name}",
        ]
    )


# if pattern is matched, the string returned by the callable will be added, the last hypothesis will be given as argument
PROOF_STRUCTURES: dict[str, tuple[str, Callable]] = {
    # by contraposition / we will prove the contrapositive / ...
    r"[Cc]ontrapos\w*": _contraposition_proof,
    # we will prove this by contradiction / absurd
    r"(?:[Cc]ontradiction|[Aa]bsurd)": _contradiction_proof,
}

SUB_STATEMENT_POSSIBILITIES: list = [Have, Induction, SimpleStatement]

space = r"\s*"


class ProofStructure(Statement):
    pattern: str = space.join(
        [
            r"",  # ignore leading space
            r"(.*)",  # anything
            f"({'|'.join(PROOF_STRUCTURES.keys())})",  # proof_structure
            r"(.*)",  # possible sub statement
            r"",  # ignore trailing space
        ]
    )

    def __init__(self, string: str, **kwargs) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}."
            )

        # for feedback
        self.match = match

        self.string = string
        self.right_side = match.group(3)

        # retrieve which proof_structure it is
        for key, translation in PROOF_STRUCTURES.items():
            if re.fullmatch(key, match.group(2)):
                self.proof_structure = key
                self.translation = translation
                break

        # possible sub-statement (e.g. when proving the contrapositive, we have ...)
        self.sub_statement = self.get_sub_statement(**kwargs)

    def get_sub_statement(self, **kwargs) -> Statement:
        for poss in SUB_STATEMENT_POSSIBILITIES:
            try:
                return poss(self.right_side, **kwargs)
            except MatchingError:
                pass

    def translate(self, **kwargs) -> str:
        if not kwargs:
            return self.translation(last_hyp="", hyp_name="hyp")

        if self.sub_statement is None:
            return self.translation(**kwargs)

        return (
            self.translation(**kwargs) + "\n\n" + self.sub_statement.translate(**kwargs)
        )

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        if self.sub_statement is None:
            sub_statement_feedback = [("ignored", self.right_side)]
        else:
            sub_statement_feedback = self.sub_statement.interpretation_feedback()

        feedback = [
            ("ignored", self.match.group(1)),
            ("keyword", self.match.group(2)),
        ]

        return feedback + sub_statement_feedback



import re

from .expression import Expression
from .algebra import Algebra
from ..utils.unfolding import unfold
from ..utils.text import indent
from ..utils.exceptions import MatchingError


class Equation(Algebra):
    pattern: str = r"(.*?)([≠=<≤>≥])(.*)"

    def __init__(self, string: str):
        # check for match
        if not re.fullmatch(self.pattern, string):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        self.string = string

        elements, separators = unfold(self.pattern, self.string)
        self.expressions: list[Expression] = [Expression(e) for e in elements]
        self.operators: list[str] = separators

    def translate(
        self, hyp_name: str = None, proof: str = None, by_calc: bool = False, **kwargs
    ) -> str:
        # sanity check
        assert not (proof is not None and by_calc), "Can't use both proof and by_calc"

        # proof
        if by_calc:
            proof = f" := by \n{indent(self.calc_block())}"
        elif proof is None:
            proof = ""
        else:
            proof = f" := {proof}"

        # hypothesis definition
        hyp_def = f"{hyp_name} : " if hyp_name else ""
        hyp = (
            hyp_def
            + self.expressions[0].translate()
            + f" {self.strongest_operator()} "
            + self.expressions[-1].translate()
        )

        return hyp + proof

    def calc_block(self) -> str:
        LINE_PROOF = f"repeat (first | ring | simp_all)"
        # keyword for the beginning
        tactic = f"calc\n"
        # 1st line
        block = f"{self.expressions[0].translate()} {self.operators[0]} {self.expressions[1].translate()}"
        # proof of 1st line
        block += f" := by {LINE_PROOF}"
        # next lines
        for expression, operator in zip(self.expressions[2:], self.operators[1:]):
            block += "\n"
            block += f"_ {operator} {expression.translate()} := by {LINE_PROOF}"

        # format complete (standalone) block
        return tactic + indent(block)

    def strongest_operator(self) -> str:
        # order of operations, from most loose to most strict
        ORDER = ["≠", ">", "<", "≥", "≤", "="]
        for op in ORDER:
            if op in self.operators:
                return op

        raise Exception(f"No operator found in {self.string}, should not get here.")



import re

from natural2lean.proof_elements.theorem.theorem import Theorem
from .statement import Statement
from .such_that import SuchThat
from .simple_statements import SimpleStatement
from ...algebra.equation import Equation
from ...propositions.multiple_propositions import MultiplePropositions
from ...propositions.implication import Implication
from ...utils.translatable import Translatable
from ...utils.exceptions import MatchingError, TranslationError
from ...utils.text import indent


"""
The system will look for matches for the keys in the whole sentence. If a match is found, the value will be passed as proof to sub-statement.
"""
SIMPLE_PROOFS: dict[str, str] = {
    r"(.*)(contradiction)(.*)": "by contradiction",
    r"(.*)(definition)(.*)": "by simp at *; assumption",
    # mod 3 is a bit specific, should extend this to allow for any modulo, but will be easier with mathlib 4
    r"(.*)(possibilities)(.*)(modulo)(.*)": "by (first | exact mod_2_poss _ | exact mod_3_poss _ | exact mod_4_poss _ | exact mod_5_poss _)",
    r"(.*)(modulo)(.*)(possibilities)(.*)": "by (first | exact mod_2_poss _ | exact mod_3_poss _ | exact mod_4_poss _ | exact mod_5_poss _)",
}


def using_theorems(theorems: list[str]) -> str:
    proof = f"repeat (first | ring | simp_all [{', '.join(theorems)}])"
    return "by \n" + indent(proof)


def induction_hypothesis(
    theorems: list[tuple[str, str, int]], **kwargs
) -> str:
    return theorems[-1][1] + " (by trivial)" * theorems[-1][2]


PARAMETRIC_PROOFS: dict[str, str] = {
    r"(.*)(induction\s*hypothesis)(.*)": induction_hypothesis,
}

DEFAULT_PROOF = "by repeat (first | trivial | ring | simp_all)"


POSSIBILITIES: list[Translatable] = [
    Implication,
    SuchThat,
    MultiplePropositions,
    SimpleStatement,
]

CALC_PROOF = 1

space = r"\s*"


class Have(Statement):
    pattern: str = space.join(
        [
            "",  # ignore leading space
            r"(.*)",  # left side
            r"([Hh]ave|[Dd]erive|[Gg]ives|[Ss]how|[Pp]rove)",  # have keyword
            r"(.*)",  # right side
            "",  # ignore trailing space
        ]
    )

    def __init__(
        self,
        string: str,
        proven_theorems: list[tuple[str, str]],
        **kwargs,
    ) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}."
            )

        # for feedback
        self.match = match

        self.string = string
        self.left_side = match.group(1).strip(" ,.;")
        self.right_side = match.group(3).strip(" ,.;")

        self.statement = self.get_statement()
        self.proof = self.get_proof(proven_theorems)

    def get_statement(self) -> Translatable:
        for poss in POSSIBILITIES:
            try:
                return poss(self.right_side)
            except MatchingError:
                continue

        raise MatchingError(
            f"Could not find a suitable substatement for have in '{self.right_side}'.\n"
        )

    def get_proof(self, proven_theorems: list[tuple[str, str, str]]) -> str:
        # checking for previously proved theorems
        used_theorems = []
        for latex_th, lean_th, _ in proven_theorems:
            if latex_th.lower() in self.string.lower():
                used_theorems.append(lean_th)

        if used_theorems:
            # return proof
            self.proof_pattern = r"(.*)(" + "|".join(used_theorems) + r")(.*)"
            return using_theorems(used_theorems)

        for pattern, param_proof in PARAMETRIC_PROOFS.items():
            if re.fullmatch(pattern, self.string):
                self.proof_pattern = pattern
                return param_proof(theorems=proven_theorems)

        # should be proved by calc
        if (
            isinstance(self.statement, MultiplePropositions)
            and len(self.statement.propositions) == 1
            and isinstance(self.statement.propositions[0], Equation)
        ):
            self.proof_pattern = r"(.*)"
            return CALC_PROOF

        # for simple proofs
        for pattern, proof in SIMPLE_PROOFS.items():
            if re.fullmatch(pattern, self.string):
                self.proof_pattern = pattern
                return proof

        # if no proof is given
        self.proof_pattern = r"(.*)"
        return DEFAULT_PROOF

    def translate(self, hyp_name=None, **kwargs) -> str:
        if hyp_name is None:
            return self.statement.translate()

        if isinstance(self.statement, SimpleStatement):
            return self.statement.translate()

        if (
            self.proof == CALC_PROOF
            and isinstance(self.statement, MultiplePropositions)
            and len(self.statement.propositions) == 1
            and isinstance(self.statement.propositions[0], Equation)
        ):
            have = f"have {self.statement.propositions[0].translate(hyp_name=hyp_name, by_calc=True)}"
            # rw = f"try rw [{hyp_name}]"
            return have # + "\n" + rw

        return f"have {self.statement.translate(hyp_name=hyp_name, proof=self.proof)}"

    def can_create_new_goals(self) -> bool:
        return self.statement.can_create_new_goals()

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        feedback = []
        # left side (no statement, can only contain a proof)
        if match := re.fullmatch(self.proof_pattern, self.left_side):
            for i, group in enumerate(match.groups()):
                group_type = "ignored" if i % 2 == 0 else "parameter"
                feedback.append((group_type, group))
        else:
            feedback.append(("ignored", self.left_side))

        # have keyword
        feedback.append(("keyword", self.match.group(2)))

        # right side (statement)
        statement_feedback = self.statement.interpretation_feedback()

        # check each section in the statement
        for section_type, section in statement_feedback:
            # if it is ignored and we have a match for our proof pattern, we add the different parts
            if section_type == "ignored" and (
                match := re.fullmatch(self.proof_pattern, section)
            ):
                for i, group in enumerate(match.groups()):
                    group_type = "ignored" if i % 2 == 0 else "parameter"
                    feedback.append((group_type, group))
            # otherwise, we add the whole section
            else:
                feedback.append((section_type, section))

        return feedback



from .separate_propositions import get_propositions
from ..utils.exceptions import MatchingError
from ..utils.translatable import Translatable
from ..algebra.identifiers import IdentifiersInSet


class MultiplePropositions(Translatable):
    def __init__(self, string: str, **kwargs) -> None:
        self.string = string

        self.propositions, self.used_keywords = get_propositions(string)

        if not self.propositions:
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

    def translate(
        self,
        hyp_name: str = None,
        separator: str = " ∧ ",
        proof: str = None,
        **kwargs,
    ) -> str:
        if len(self.propositions) == 1:
            return self.propositions[0].translate(
                hyp_name=hyp_name, proof=proof, **kwargs
            )

        hyp_def = f"{hyp_name} : " if hyp_name else ""
        props = separator.join([prop.translate() for prop in self.propositions])
        proof = f" := {proof}" if proof else ""

        return hyp_def + props + proof

    def get_identifiers(self) -> list[IdentifiersInSet]:
        return [
            prop for prop in self.propositions if isinstance(prop, IdentifiersInSet)
        ]

    def get_non_identifiers(self) -> list[Translatable]:
        return [
            prop for prop in self.propositions if not isinstance(prop, IdentifiersInSet)
        ]

    def translate_identifiers(self, separator: str = " ") -> str:
        return separator.join([prop.translate() for prop in self.get_identifiers()])

    def translate_non_identifiers(self, separator: str = " ∧ ") -> str:
        return separator.join([prop.translate() for prop in self.get_non_identifiers()])

    def can_create_new_goals(self) -> bool:
        return any(prop.can_create_new_goals() for prop in self.propositions)

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        if not self.used_keywords:
            return [("ignored", self.string)]
        
        # check for all keywords
        keyword_positions = []
        for kw in self.used_keywords:
            if (pos := self.string.find(kw)) != -1:
                keyword_positions.append((pos, kw))
        
        # sort by position
        keyword_positions = sorted(keyword_positions)
        
        # separate into ignored and parameters
        result = []
        current_position = 0
        
        for pos, kw in keyword_positions:
            if pos > current_position:
                result.append(("ignored", self.string[current_position:pos]))
            result.append(("parameter", kw))
            current_position = pos + len(kw)
        
        result.append(("ignored", self.string[current_position:]))
        
        return result
        
        



import re
from .statement import Statement
from ...propositions.multiple_propositions import MultiplePropositions
from ...algebra.identifiers import Identifier
from ...utils.exceptions import MatchingError, TranslationError


space = r"\s*"


class SuchThat(Statement):
    pattern: str = space.join(
        [
            "",  # ignore leading space
            r"(.*)",  # left side
            r"([Ss]uch)",  # such that keyword
            r"([Tt]hat)",  # with space between
            r"(.*)",  # right side
            "",  # ignore trailing space
        ]
    )

    def __init__(self, string, **kwargs) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        # for feedback
        self.match = match

        # identifiers
        self.identifiers = MultiplePropositions(match.group(1))

        if len(self.identifiers.get_identifiers()) == 0:
            raise TranslationError(
                f"Found no identifiers before 'such that' in '{string}'.\n"
            )

        if len(self.identifiers.get_non_identifiers()) > 0:
            raise TranslationError(
                f"A proposition before 'such that' in '{string}' is not an identifier definition.\n"
            )

        # hypotheses
        self.hypotheses = MultiplePropositions(match.group(4))

    def translate(self, hyp_name=None, proof=None, **kwargs) -> str:
        assert (proof is None) == (
            hyp_name is None
        ), "Should always use proof with hyp_name."

        identifiers: list[Identifier] = []
        for ident_in_set in self.identifiers.get_identifiers():
            identifiers += ident_in_set.identifiers

        id_names = [i.translate() for i in identifiers]
        id_defs = " ".join(
            [idents.translate() for idents in self.identifiers.get_identifiers()]
        )
        hyp_def = self.hypotheses.translate()

        constructor = f"⟨{', '.join(id_names)}, {hyp_name}⟩"
        exists = f"∃ {id_defs}, {hyp_def}"

        return f"{constructor} : {exists} := {proof}"

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        feedback = [
            ("keyword", self.match.group(2)),
            ("keyword", self.match.group(3)),
        ]

        return (
            self.identifiers.interpretation_feedback()
            + feedback
            + self.hypotheses.interpretation_feedback()
        )



import re
from .multiple_propositions import MultiplePropositions
from ..utils.exceptions import MatchingError
from ..utils.translatable import Translatable

space = r"\s*"
punctuation = " .,;:"


class Implication(Translatable):
    pattern: str = space.join(
        [
            "",  # ignore leading space
            r"([Ii]f)",  # if keyword
            r"(.*)",  # hypotheses
            r"([Tt]hen)",  # then keyword
            r"(.*)",  # theses
            "",  # ignore trailing space
        ]
    )

    def __init__(self, string: str) -> None:
        match = re.fullmatch(self.pattern, string)
        if not match:
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        # for feedback
        self.match = match

        self.string = string
        self.hypotheses = MultiplePropositions(match.group(2).strip(punctuation))
        self.theses = MultiplePropositions(match.group(4).strip(punctuation))

    def translate(self, hyp_name: str = None, proof: str = None, **kwargs) -> str:
        assert (
            proof is None == hyp_name is None
        ), "Should always use proof with hyp_name."

        hyp_def = f"{hyp_name} : " if hyp_name else ""
        proof = f" := {proof}" if proof else ""

        hypotheses_tr = self.hypotheses.translate(separator=" → ")
        theses_tr = self.theses.translate()

        implication = f"{hypotheses_tr} → {theses_tr}"

        return hyp_def + implication + proof

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        return [
            ("keyword", self.match.group(1)),
            *self.hypotheses.interpretation_feedback(),
            ("keyword", self.match.group(3)),
            *self.theses.interpretation_feedback(),
        ]



import re
from .statement import Statement
from ...utils.exceptions import MatchingError


"""
Simple statement such as "we have a contradiction". If the key is matched, the value will be returned by translate. You should consider that the value should already be in tactic mode.
"""
SIMPLE_STATEMENTS: dict[str, str] = {
    r"contradiction": "contradiction",
    r"trivial": "repeat (first | trivial | ring | simp_all)",
}

space = r"\s*"


class SimpleStatement(Statement):
    pattern: str = space.join(
        [
            r"",  # ignore leading space
            r"(.*)",  # anything
            f"({'|'.join(SIMPLE_STATEMENTS.keys())})",  # statement
            r"(.*)",  # anything
            r"",  # ignore trailing space
        ]
    )

    def __init__(self, string, **kwargs) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}."
            )

        # for feedback
        self.match = match

        self.string = string

        # retrieve which proof_structure it is
        for key, translation in SIMPLE_STATEMENTS.items():
            if re.fullmatch(key, match.group(2)):
                self.statement = key
                self.translation = translation
                break

    def translate(self, **kwargs) -> str:
        return self.translation

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        feedback = [
            ("ignored", self.match.group(1)),
            ("keyword", self.match.group(2)),
            ("ignored", self.match.group(3)),
        ]

        return feedback



import re
from .have import Have
from .induction import Induction
from .proof_structure import ProofStructure
from .simple_statements import SimpleStatement
from ...algebra.equation import Equation
from ...algebra import get_algebra
from ...utils.exceptions import MatchingError
from .statement import Statement

space = r"\s*"

SUB_STATEMENT_POSSIBILITIES = [
    Induction,
    Have,
    ProofStructure,
    SimpleStatement,
]


class InductionCase(Statement):
    """Case will find an equation in the sentence and compare the right side with the variable on which induction is being performed. If they match, it will create the case and check for a possible sub-statement. Otherwise it will raise a MatchingError."""

    pattern: str = space.join(
        [
            r"",  # ignore leading space
            r"(.*?)",  # anything ("for", ...)
            r"(\$.+?\$)",  # the variable ("$n = 0$", ...)
            r"(.*)",  # potential sub-statement
            r"",  # ignore trailing space
        ]
    )

    def __init__(self, string, proof_status: list[Statement], **kwargs) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match '{string}' in {self.__class__.__name__}."
            )

        # for feedback
        self.match = match

        self.string = string
        self.right_side = match.group(3)

        # find the first algebraic expression in the string
        alg = get_algebra(match.group(2))

        if isinstance(alg, Equation) and len(alg.expressions) == 2:
            self.variable = alg.expressions[0]
            self.expression = alg.expressions[1]
        else:
            raise MatchingError(
                f"The first equation does not correspond to a specific case (such as $n=1$)."
            )

        for statement in proof_status:
            if (
                isinstance(statement, Induction)
                and statement.variable.translate() == self.variable.translate()
            ):
                self.induction = statement
                break
        else:
            # did not find an induction in the proof_status on a corresponding variable
            raise MatchingError(
                f"Could not find an induction with '{self.variable.string}' in the previous statements."
            )

        self.sub_statement = self.get_sub_statement(**kwargs)

    def get_sub_statement(self, **kwargs) -> Statement:
        for poss in SUB_STATEMENT_POSSIBILITIES:
            try:
                return poss(self.right_side, **kwargs)
            except MatchingError:
                pass

    def translate(self, **kwargs) -> str:
        if self.sub_statement:
            sub_tr = self.sub_statement.translate(**kwargs)
        else:
            sub_tr = ""

        return f"| {self.expression.translate()} =>\nskip\n{sub_tr}"

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        if self.sub_statement is None:
            sub_statement_feedback = [("ignored", self.right_side)]
        else:
            sub_statement_feedback = self.sub_statement.interpretation_feedback()

        feedback = [
            ("ignored", self.match.group(1)),
            ("parameter", self.match.group(2)),
        ]

        return feedback + sub_statement_feedback



from .example import Example
from .named_theorem import NamedTheorem
from .theorem import Theorem
from ...utils.exceptions import MatchingError, TranslationError

POSSIBILITIES = [NamedTheorem, Example]


def get_theorem(string: str, **kwargs) -> Theorem:
    for poss in POSSIBILITIES:
        try:
            return poss(string, **kwargs)
        except MatchingError:
            pass

    raise TranslationError(
        f"Could not match any theorem for '{string}', tried "
        + ", ".join([p.__name__ for p in POSSIBILITIES])
        + ".\n"
    )



import re
from typing import Iterable
from .function import Function
from .proposition_constants import (
    SEPARATORS,
    NEGATIONS,
    FUNCTIONS,
    VALIDITY_CHECKS,
    apply_replacements,
)
from ..algebra import (
    get_algebra,
    Algebra,
    Equation,
    Expression,
    IdentifiersInSet,
    MultipleIdentifiers,
)
from ..utils.exceptions import MatchingError, TranslationError
from ..utils.translatable import Translatable
from ..algebra.translation_constants import SETS, MathSet
from ..algebra.expression_possibilities import ExpressionPossibilities


# ------------------ MAIN FUNCTIONS ------------------


def get_propositions(string: str) -> tuple[list[Translatable], set[str]]:
    props = []
    keywords = set()
    for prop, kwds in separate_propositions(string):
        props.append(prop)
        keywords.update(kwds)
    return props, keywords


def separate_propositions(string: str) -> Iterable[tuple[Translatable, list[str]]]:
    """Separates multiple propositions (PropA, PropB and PropC).

    Args:
        string (str): the input string
    """
    # also yield the last proposition ;)
    separator_spans = [(len(string), len(string))]

    # get all the spans for the separators
    for separator in SEPARATORS:
        separator_spans += [m.span() for m in re.finditer(separator, string)]
    separator_spans = sorted(separator_spans)

    # yield different propositions
    last_stop = 0
    for start, stop in separator_spans:
        if is_valid(string[last_stop:start]):
            for proposition, keywords in split_proposition(string[last_stop:start]):
                yield proposition, keywords
            last_stop = stop


def split_proposition(string: str) -> Iterable[tuple[Translatable, list[str]]]:
    """Splits a proposition ($a$ and $b$ are even natural numbers) into multiple propositions ($a \\in \\mathbb{N}$, $b \\in \\mathbb{N}$). Also returns a list of keywords that have been used.

    Args:
        string (str): input proposition
    """
    string = apply_replacements(string)

    # match math mode
    if (match := re.fullmatch("(.*)(\$.+\$)(.*)", string)) is not None:
        try:
            math = get_algebra(match.group(2))
        except MatchingError:
            math = None

        # set
        if isinstance(math, MultipleIdentifiers):
            set_, kw = get_set(string)
            if set_:
                if isinstance(math, IdentifiersInSet):
                    raise TranslationError(
                        f"Multiple sets in the same proposition; found '{set_}' and '{math.set.latex}' in '{string}'\n"
                    )

                algebra = get_algebra("$" + math.string + " \\in " + set_.latex + "$")
                assert isinstance(algebra, IdentifiersInSet)
                yield algebra, [kw, match.group(2)]

        # multiple identifiers in set directly in the proposition
        if isinstance(math, IdentifiersInSet):
            yield math, [match.group(2)]

        # functions on identifiers or expressions
        if (
            isinstance(math, MultipleIdentifiers)
            or isinstance(math, Expression)
            or isinstance(math, ExpressionPossibilities)
        ):
            for function, keywords in get_functions(string):
                yield function, keywords + [match.group(2)]

        # equation (should not have a function associated to it)
        if isinstance(math, Equation) or isinstance(math, ExpressionPossibilities):
            yield math, [match.group(2)]


# ------------------ EXTRACT INFORMATION ------------------


def get_set(string: str) -> tuple[MathSet, str]:
    """Extracts the set associated to the proposition.

    Args:
        string (str): the input string
    """
    string = string.lower()
    matched_set = None
    keyword = None
    for set_ in SETS.values():
        for word in set_.full_words:
            if word in string:
                if matched_set != None:
                    raise TranslationError(
                        f"Multiple sets in the same proposition; found '{keyword}' and '{word}' in '{string}'\n"
                    )
                matched_set = set_
                keyword = word

    return matched_set, keyword


def get_functions(string: str) -> Iterable[tuple[Function, list[str]]]:
    """Extracts the functions associated to the proposition.

    Args:
        string (str): the input string
    """
    for name, pattern, order_arguments in FUNCTIONS:
        # skip if no match
        if (match := re.search(pattern, string)) is None:
            continue

        # keywords (all groups)
        keywords = list(match.groups())
        # negate function
        neg = get_negation(string)
        if neg is not None:
            keywords.append(neg)
            name = f"¬ {name}"
        # match arguments (even groups)
        args: list[Algebra] = [get_algebra(arg) for arg in match.groups()[::2]]

        # first group can be multiple identifiers
        if isinstance(args[0], MultipleIdentifiers):
            for identifier in args[0].identifiers:
                yield Function(
                    name, order_arguments(*[identifier, *args[1:]])
                ), keywords
            continue

        if isinstance(args[0], Expression):
            yield Function(name, order_arguments(*args)), keywords
            continue


# ---------------- SMALL UTILITY FUNCTIONS ----------------


def is_valid(string: str):
    """Checks if the string is valid (according to VALIDITY_CHECKS).

    Args:
        string (str): the input string
    """
    for validity_check in VALIDITY_CHECKS:
        if not validity_check(string):
            return False
    return True


def get_negation(string: str) -> str:
    """Returns the negation if it is present in the string, None otherwise.

    Args:
        string (str): the input string
    """
    string = string.lower()
    for neg in NEGATIONS:
        if neg in string:
            return neg
    return None



import os
import shutil
import platform
from pathlib import Path, PureWindowsPath

from natural2lean.utils.translatable import Translatable
from .proof_elements.theorem import get_theorem
from .proof_elements.theorem.theorem import Theorem
from .proof_elements.statement import CCL_POSSIBILITIES, get_statement
from .proof_elements.statement.statement import Statement
from .utils.text import indent, subscript
from .utils.exceptions import LeanError, NoConclusion
from .lean_interaction.conclude_proof import get_conclusion
from .lean_interaction.lean_feedback import State, lean_feedback

DEFAULT_PATHS = {
    "Darwin": Path.home() / Path(".natural2lean"),
    "Linux": Path.home() / Path(".natural2lean"),
    "Windows": Path.home() / Path("AppData/Roaming/.natural2lean"),
}

LEAN_PROJECT_GIT_REPO = (
    "https://github.com/Augustindou/natural2lean-lean-project-template.git"
)
GIT_LOCATION = ".git"

LEAN_HEADER = "\n".join(
    [
        "import LeanUtils",
        'def main : IO Unit := IO.println s!"Hello, world!"',
        "open Nat",
    ]
)


windows = platform.system() == "Windows"


class Translator:
    """An object managing a stack of theorems and statements that form proofs."""

    def __init__(self, lean_project_directory: str = None):
        self.stack: list[State] = [
            State(goals=[], last_statement=None, lean_text=LEAN_HEADER)
        ]
        # containing latex name, lean name, number of arguments
        self.theorems: list[tuple[str, str, int]] = []
        # position in proof (for now, only inductions)
        self.proof_status: list[Statement] = []
        # last statement that failed to be translated
        self.last_failed_statement: Translatable = None

        try:
            default_path = DEFAULT_PATHS[platform.system()]
        except KeyError:
            raise Exception(
                f"Unsupported platform: `{platform.system()}`, please report this bug."
            )

        # path to project
        if lean_project_directory:
            if windows:
                lean_project_directory = PureWindowsPath(lean_project_directory)
            self.project_directory = Path(lean_project_directory)
        else:
            self.project_directory = default_path

        # check if project exists
        if not self.project_directory.exists():
            # if exists at the default location, copy the folder
            if default_path.exists():
                shutil.copytree(
                    default_path,
                    self.project_directory,
                )
            # otherwise, download the project
            else:
                self.project_directory.mkdir(parents=True)
                print("Downloading lean project template...\n")
                os.system(f"git clone {LEAN_PROJECT_GIT_REPO} {self.project_directory}")
                print()

        # lake build to make sure lean it will work (this will download Mathlib if it is not already)
        try:
            lean_feedback(LEAN_HEADER, None, self.project_directory)
        except LeanError:
            raise Exception(
                "The lean header alone should not cause an error, please check that lean4 is installed on your system, and report this bug if it is."
            )

    def new(self, string: str) -> State:
        """Adds a new state to the stack, this function will call new_theorem or new_statement depending on whether the last state has goals to solve or not.

        Args:
            string (str): the string to be parsed

        Returns:
            State: the state after the theorem or statement has been added

        Raises:
            AssertionError: if the last state still has a goal to be solved.
            TranslationError: if the string could not be parsed into a theorem.
            LeanError: if the translation by the system could not be understood by lean.
        """
        old_state: State = self.state()
        if len(old_state.goals) == 0:
            return self.new_theorem(string)
        else:
            return self.new_statement(string)

    def new_theorem(self, string: str) -> State:
        """Adds a new state to the stack, with a new theorem. This will create goals to solve, and these goals will be part of the returned state.

        Args:
            string (str): the string to be parsed into a theorem

        Returns:
            State: the state after the theorem has been added

        Raises:
            AssertionError: if the last state still has a goal to be solved.
            TranslationError: if the string could not be parsed into a theorem.
            LeanError: if the translation by the system could not be understood by lean.
        """
        theorem = get_theorem(string, i=len(self.theorems))

        self.proof_status = []

        old_state: State = self.state()
        assert (
            len(old_state.goals) == 0
        ), "Should close goals before proving new theorem"

        lean_text = old_state.lean_text + "\n\n" + theorem.translate()

        # lean_feedback(lean_text) should throw an error if lean could not understand the translated theorem
        try:
            new_state = State(
                goals=lean_feedback(lean_text, theorem, self.project_directory),
                last_statement=theorem,
                lean_text=lean_text,
            )
        except LeanError as e:
            self.last_failed_statement = theorem
            raise e

        # add theorem to list
        if isinstance(theorem, Theorem):
            self.theorems.append(
                (theorem.latex_name, theorem.lean_name, theorem.n_args)
            )

        self.stack.append(new_state)
        return new_state

    def new_statement(self, string: str) -> State:
        """Adds a new state to the stack, with a new statement. The stack should have at least one goal. This will add hypotheses to the goal, these hypotheses can be retrieved from the returned state.

        Args:
            string (str): the string to be parsed into a statement

        Returns:
            State: the state after the statement has been added

        Raises:
            AssertionError: if the last state doesn't have any goal to be solved.
            TranslationError: if the string could not be parsed into a theorem.
            LeanError: if the translation by the system could not be understood by lean.
        """
        statement = get_statement(
            string, proven_theorems=self.theorems, proof_status=self.proof_status
        )

        if statement.change_status():
            self.proof_status.append(statement)

        old_state: State = self.state()
        assert (
            len(old_state.goals) > 0
        ), "Should start a theorem before adding statements"

        try:
            lean_fb, translation = get_conclusion(
                state=old_state,
                goals=self._all_current_goals(),
                statement=statement,
                project_directory=self.project_directory,
            )
        except NoConclusion:
            if any(isinstance(statement, poss) for poss in CCL_POSSIBILITIES):
                self.last_failed_statement = statement
                raise NoConclusion(
                    f"Could not match a non-conclusive statement, nor conclude a proof with '{string}'.\n"
                )

            hyp_name = f"h{subscript(len(old_state.goals[0].hypotheses))}"
            if old_state.goals[0].hypotheses:
                last_hyp = old_state.goals[0].hypotheses[-1][0]
            else:
                last_hyp = None
            translation = statement.translate(
                hyp_name=hyp_name,
                last_hyp=last_hyp,
            )

            try:
                lean_fb = lean_feedback(
                    input=old_state.lean_text + "\n\n" + indent(translation),
                    statement=statement,
                    project_directory=self.project_directory,
                )
            except LeanError as e:
                self.last_failed_statement = statement
                raise e

        lean_text = old_state.lean_text + "\n\n" + indent(translation)

        # lean_feedback(lean_text) should have thrown an error if lean could not understand the translated theorem
        new_state = State(
            goals=lean_fb,
            last_statement=statement,
            lean_text=lean_text,
        )

        if (
            len(new_state.goals) > len(old_state.goals)
            and not statement.can_create_new_goals()
        ):
            self.last_failed_statement = statement
            raise LeanError("Lean detected an error in your statement.\n")

        self.stack.append(new_state)
        return new_state

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        """Returns a list of tuples (`type`, `section`) describing the way the last input was processed.

        The type can be one of the following:
        - `"keyword"`: the `section` of the string is a keyword used to interpret the string (e.g. "theorem", "have")
        - `"parameter"`: the `section` of the string is a parameter used in the translation (e.g. a theorem name, a math expression)
        - `"ignored"`: the `section` of the string was ignored (most filling words like "we", "therefore", ...)
        """
        if self._is_bottom_state():
            return None
        return self.stack[-1].last_statement.interpretation_feedback()

    def failed_statement_interpretation(self) -> list[tuple[str, str]]:
        if self.last_failed_statement is None:
            return None
        return self.last_failed_statement.interpretation_feedback()

    def backtrack(self) -> State:
        """Removes the last input given to the Translator, hence, the last state of the stack. If no input had been given earlier, this function will return None.

        Returns:
            State: the state after the last input has been removed.
        """
        if self._is_bottom_state():
            return None

        # remove last state
        old_state = self.stack.pop()

        # remove theorem if it was popped
        if isinstance(old_state.last_statement, Theorem):
            self.theorems.pop()

        return self.state()

    def state(self) -> State:
        """Returns the current state of the Translator.

        Returns:
            State: the current state of the Translator.
        """
        return self.stack[-1]

    def _all_current_goals(self) -> set[str]:
        current_goals = set()
        for state in self.stack[::-1]:
            # stop when exiting the current proof
            if not state.goals:
                break
            # add each goal
            current_goals.add(state.goals[0].goal)

        return current_goals

    def _is_bottom_state(self) -> bool:
        return len(self.stack) == 1



import re
from .theorem import Theorem, STATEMENT_POSSIBILITIES
from ...utils.text import subscript
from ...utils.exceptions import MatchingError

space = r"\s*"


class Example(Theorem):
    pattern: str = space.join(
        [
            "",  # ignore leading space
            r"(.*)",  # Theorem statement
            "",  # ignore trailing space
        ]
    )

    def __init__(self, string: str, i: int) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        self.latex_name = f"th{i}"
        self.lean_name = f"th{i}"

        # content
        for poss in STATEMENT_POSSIBILITIES:
            try:
                self.statement = poss(match.group(1))
                self.set_hypotheses_and_theses()
                return
            except MatchingError:
                pass

        raise MatchingError(
            f"Could not match any statement to prove for {string}, tried {', '.join([p.__name__ for p in STATEMENT_POSSIBILITIES])}."
        )

    def interpretation_feedback(self) -> list[tuple[str, str]]:

        return self.statement.interpretation_feedback()

    # translate method is inherited



from ...utils.translatable import Translatable
from ...propositions.multiple_propositions import MultiplePropositions
from ...propositions.implication import Implication
from ...utils.text import subscript

STATEMENT_POSSIBILITIES = [Implication, MultiplePropositions]


class Theorem(Translatable):
    def __init__(self):
        self.statement: Translatable = None
        self.latex_name: str = None
        self.lean_name: str = None
        self.n_args: int = None
        raise NotImplementedError

    def set_hypotheses_and_theses(self):
        if isinstance(self.statement, Implication):
            self.hypotheses = self.statement.hypotheses
            self.theses = self.statement.theses
            self.n_args = len(self.hypotheses.propositions)
        elif isinstance(self.statement, MultiplePropositions):
            self.hypotheses = None
            self.theses = self.statement
            self.n_args = len(self.theses.get_identifiers())

    def translate(self, **kwargs) -> str:
        if self.hypotheses:
            lean_identifiers = self.hypotheses.translate_identifiers()
            lean_hypotheses = " ".join(
                [
                    f"(h{subscript(i)} : {h.translate()})"
                    for i, h in enumerate(self.hypotheses.get_non_identifiers())
                ]
            )
            lean_theses = self.theses.translate()

            return f"theorem {self.lean_name} {lean_identifiers} {lean_hypotheses} : {lean_theses} := by"

        else:
            lean_identifiers = self.theses.translate_identifiers()
            lean_theses = self.theses.translate_non_identifiers()

            return f"theorem {self.lean_name} {lean_identifiers} : {lean_theses} := by"



import re
from ...algebra import get_algebra
from ...algebra.identifiers import MultipleIdentifiers
from ...utils.exceptions import MatchingError
from .statement import Statement


space = r"\s*"


class Induction(Statement):
    pattern: str = space.join(
        [
            r"",  # ignore leading space
            r"(.*)",  # anything
            r"(induction)",  # induction keyword
            r"(.*?)",  # anything
            r"(\$.+\$)",  # the variable
            r"(.*)",  # anything
            r"",  # ignore trailing space
        ]
    )

    def __init__(self, string: str, **kwargs) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match '{string}' in {self.__class__.__name__}."
            )

        # for feedback
        self.match = match

        self.string = string
        expr = get_algebra(match.group(4))

        if isinstance(expr, MultipleIdentifiers) and len(expr.identifiers) == 1:
            self.variable = expr.identifiers[0]
        else:
            raise MatchingError(
                f"Matched {expr.__class__.__name__} in '{self.string}' instead of single identifier."
            )

    def translate(self, **kwargs) -> str:
        return f"match {self.variable.translate()} with"

    def change_status(self) -> bool:
        return True

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        feedback = [
            ("ignored", self.match.group(1)),
            ("keyword", self.match.group(2)),  # "induction"
            ("ignored", self.match.group(3)),
            ("parameter", self.match.group(4)),  # the variable
            ("ignored", self.match.group(5)),
        ]

        return feedback



from __future__ import annotations
from .algebra import Algebra
from .equation import Equation
from .expression import Expression
from .expression_possibilities import ExpressionPossibilities
from .identifiers import IdentifiersInSet, MultipleIdentifiers
from ..utils.exceptions import MatchingError, TranslationError

POSSIBILITIES = [
    Equation,
    ExpressionPossibilities,
    IdentifiersInSet,
    MultipleIdentifiers,
    Expression,
]


def get_algebra(latex_string: str, match_type: str = "full") -> Algebra:
    algebra = Algebra(latex_string, match_type=match_type)

    for poss in POSSIBILITIES:
        try:
            return poss(algebra.lean_string)
        except MatchingError:
            pass

    raise TranslationError(
        f"Could not match any algebra for '{latex_string}', tried "
        + ", ".join([p.__name__ for p in POSSIBILITIES])
        + ".\n"
    )



from pathlib import Path
import re

from natural2lean.proof_elements.statement.statement import Statement
from .lean_feedback import LeanBlock, State, lean_feedback
from ..utils.text import indent
from ..utils.exceptions import LeanError, NoConclusion
from ..utils.translatable import Translatable


# if any element of the key is in the goal, the system will add the value to the proof if it solves a goal
CONCLUSIONS: dict[tuple[str], str] = {
    (r"¬divisible", r"even", r"odd", r"divisible"): "simp_all",
    (r"∃ .+, .+",): "exact ⟨_, by repeat (first | trivial | ring | simp_all)⟩",
    (
        r"even",
    ): "apply even_rewrite.mpr; exact ⟨_, by repeat (first | trivial | ring | simp_all)⟩",
    (
        r"odd",
    ): "apply odd_rewrite.mpr; exact ⟨_, by repeat (first | trivial | ring | simp_all)⟩",
    (
        r"%.*=",
        r"divisible",
    ): "apply mod_rewrite.mpr; exact ⟨_, by repeat (first | trivial | ring | simp_all)⟩",
    (r"",): "assumption",
}


def get_conclusion(
    state: State,
    goals: str,
    statement: Translatable,
    project_directory: Path,
) -> tuple[list[LeanBlock], str]:
    """Returns the conclusion to add to the proof if the statement is a goal, or None if the conclusion does not work / the statement is not a goal."""
    if statement_is_goal(statement, goals):
        return find_conclusion(state, goals, statement, project_directory)
    raise NoConclusion(f"No conclusion found for {state.goals[0].goal}")


def statement_is_goal(statement: Translatable, goals: set[str]) -> bool:
    # TODO : improve the method for this
    # get first line
    tr = statement.translate().splitlines()[0]

    # remove parenth and spaces
    tr = tr.replace("(", "").replace(")", "").replace(" ", "")

    for goal in goals:
        # remove parenth and spaces
        goal = goal.replace("(", "").replace(")", "").replace(" ", "")
        if goal in tr:
            return True
    return False


def find_conclusion(
    state: State, goals: list[str], statement: Statement, project_directory: Path
) -> tuple[list[LeanBlock], str]:
    for indicators, ccl in CONCLUSIONS.items():
        indented_ccl = indent(ccl.strip())
        for indicator in indicators:
            if any(re.search(indicator, goal) for goal in goals):
                try:
                    lean_fb = lean_feedback(
                        input=state.lean_text + "\n\n" + indented_ccl,
                        statement=statement,
                        project_directory=project_directory,
                    )
                except LeanError:
                    break

                if len(lean_fb) < len(state.goals):
                    return lean_fb, ccl

                break
    raise NoConclusion(f"No conclusion found for {state.goals[0].goal}")



import re
from .expression import Expression
from .algebra import Algebra
from ..utils.exceptions import MatchingError
from ..utils.unfolding import unfold

space = r"\s*"


class ExpressionPossibilities(Algebra):
    multiple_expressions_pattern = f"(.*?)" + r"(?:(,)(.*))?"
    pattern: str = space.join(
        [
            "",  # ignore leading space
            f"({Expression.pattern})",
            r"∈",
            r"\[",
            f"({multiple_expressions_pattern})",
            r"\]",
            "",  # ignore trailing space
        ]
    )

    def __init__(self, string: str) -> None:
        # check for match
        match = re.fullmatch(self.pattern, string)
        if not match:
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        self.string = string.strip()

        # get expression
        self.expression: Expression = Expression(match.group(1))

        # get possibilities
        elements, separators = unfold(self.multiple_expressions_pattern, match.group(2))
        self.possibilities: list[Expression] = [Expression(e) for e in elements]

    def translate(self, hyp_name=None, proof=None, **kwargs) -> str:
        assert (proof is None) == (
            hyp_name is None
        ), "Should always use proof with hyp_name."

        disjunction = " ∨ ".join(
            [
                f"{self.expression.translate()} = {p.translate()}"
                for p in self.possibilities
            ]
        )

        if hyp_name is None:
            return disjunction

        hyp_def = f"{hyp_name} : "
        proof = f" := {proof}"

        hyp = hyp_def + disjunction + proof
        split_goals = f"rcases {hyp_name} with " + " | ".join(
            [hyp_name] * len(self.possibilities)
        )

        return hyp + "\n" + split_goals

    def can_create_new_goals(self) -> bool:
        return True



class LeanError(Exception):
    """
    Exception when lean failed.
    """


class NoConclusion(ValueError):
    """
    Exception when a conclusion to a goal could not be found.
    """


class TranslationError(ValueError):
    """
    Exception when translation failed (system could not translate a statement).
    """


class MatchingError(ValueError):
    """
    Exception when matching failed (system could not match the string).
    """



from natural2lean import Translator
# end-to-end tests can be found in the natural2lean-cli repo at https://github.com/Augustindou/natural2lean-cli

from natural2lean.utils.exceptions import LeanError


translator = Translator()

translator.new(r"Theorem Square of even number is even: if $m \in \mathbb{N}$ is even, then $m^2$ is even.")
translator.new(r"By definition, if $m$ is even, we have a natural number $n$ such that $m = 2n$.")
try:
    translator.new(r"Therefore we have $m^2 = (2n)^3 = 4n^2 = 2(2n^2)$.")
except LeanError as e:
    translator.failed_statement_interpretation()


import re


def unfold(pattern: str, string: str) -> tuple[list[str], list[str]]:
    """Recursively unfolds the pattern in the string.

    Args:
        pattern (str): pattern to unfold, must contain 4 groups : complete match, element, separator, last element or tail (to be matched recursively).
        string (str): string to unfold.

    Returns:
        elements (list[str]): list of (n) elements in the string.
        separators (list[str]): list of (n-1) separators between elements.
    """
    elements = []
    separators = []

    # first match, has to work
    match = re.fullmatch(pattern, string)
    if match == None:
        return None, None

    # adding the first element
    elements.append(match.group(1).strip())

    # case for no second part (only 1 element)
    if match.group(3) == None:
        return elements, separators

    # adding the separator
    separators.append(match.group(2).strip())

    # matching the second part
    if re.fullmatch(pattern, match.group(3)) == None:
        # no more unfolding
        elements.append(match.group(3).strip())
        return elements, separators

    else:
        # unfold recursively
        new_elements, new_separators = unfold(pattern, match.group(3))
        # adding at the end of the lists
        elements += new_elements
        separators += new_separators

        return elements, separators


if __name__ == "__main__":
    pattern = r"( *([a-zA-Z]\w*) *(,) *([a-zA-Z]\w*(?: *, *[a-zA-Z]\w*)*) *)"
    print(unfold(pattern, "x_1 , x_2 , x_3"))



import re
from .algebra import Algebra
from .translation_constants import SETS
from ..utils.unfolding import unfold
from ..utils.exceptions import MatchingError


class Identifier(Algebra):
    # starts with a letter and then anything alphanumeric, no spaces inside but allow spaces around
    pattern: str = r"\s*[a-zA-Z]\w*\s*"

    def __init__(self, string: str) -> None:
        # check for match
        if not re.fullmatch(self.pattern, string):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        self.string = string.strip()

    def translate(self, **kwargs) -> str:
        return self.string


class MultipleIdentifiers(Algebra):
    # start with an identifier possibly followed by a comma, then anything
    # this will allow for only one identifier
    pattern: str = f"({Identifier.pattern})" + r"(?:(,)(.*))?"

    def __init__(self, string: str) -> None:
        # check for match
        if not re.fullmatch(self.pattern, string):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        self.string = string.strip()

        # get identifiers
        elements, separators = unfold(self.pattern, self.string)
        # an error will be raised here if there is a problem with an identifier
        self.identifiers: list[Identifier] = [Identifier(e) for e in elements]

    def translate(self, **kwargs) -> str:
        return " ".join([i.translate() for i in self.identifiers])


class IdentifiersInSet(MultipleIdentifiers):
    pattern: str = f"({MultipleIdentifiers.pattern})" + r"∈\s*(.+)"

    def __init__(self, string: str) -> None:
        # check for match
        match = re.fullmatch(self.pattern, string)
        if not match:
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        self.string = string.strip()

        self.identifiers = MultipleIdentifiers(match.group(1)).identifiers
        self.set = SETS[match.group(5).strip()]

    def translate(self, **kwargs) -> str:
        lean_identifiers = " ".join([i.translate() for i in self.identifiers])

        return f"({lean_identifiers} : {self.set.lean})"



import re
from ..utils.exceptions import TranslationError
from .translation_constants import (
    START_OF_FUNCTION,
    END_OF_FUNCTION,
    LATEX_SYMBOLS,
    LATEX_FUNCTIONS,
    IMPLICIT_OPERATIONS,
    LatexFunction,
)


class Latex2LeanMath:
    """Translation from LaTeX math (`a\\frac{a + 1}{b^2}`) to lean4 math (`a * ((a+1) / (b^2))`) can be done by calling `str(Latex2LeanMath(latex_string))` or simply `translate_latex_math(latex_string)`"""

    def __init__(self, latex_string: str) -> None:
        self.latex_string = latex_string
        self.result_string = ""
        self.symbol_replacement()
        self.function_replacement()
        self.implicit_operations()

    def symbol_replacement(self):
        for k, v in LATEX_SYMBOLS.items():
            self.latex_string = self.latex_string.replace(k, v)

    def function_replacement(self, position: int = 0) -> int:
        next_function_start, next_function_end, next_function = self.get_next_function(
            position
        )
        next_start_start, next_start_end = self.get_next_start(position)
        next_end_start, next_end_end = self.get_next_end(position)

        # no function or arg
        if (
            next_function == None
            and next_start_start == None
            and next_end_start == None
        ):
            self.result_string += self.latex_string[position:]
            return -1

        # function without end
        if next_end_start == None:
            raise TranslationError(
                f"LaTeX function without end"
                + f'\n        in "{self.latex_string}"'
                + f'\n        at "{self.latex_string[next_function_start:]}"'
                + "\n"
            )

        # end found before start
        if next_start_start == None or next_end_start < next_start_start:
            self.result_string += self.latex_string[position:next_end_start]
            return next_end_end

        # function found
        if next_start_start < next_end_start and next_function_start < next_start_start:
            # add part before function
            self.result_string += self.latex_string[position:next_function_start]

            for i in range(len(next_function.separators) - 1):
                # go to next argument
                next_start_start, next_start_end = self.get_next_start(position)
                next_end_start, next_end_end = self.get_next_end(position)

                # if some argument is missing START_OF_FUNCTION
                if next_start_start == None:
                    raise TranslationError(
                        f"LaTeX function with wrong number of arguments"
                        + f'\n        in "{self.latex_string}"'
                        + f'\n        at "{self.latex_string[next_function_start:]}"'
                        + "\n"
                    )

                # add intermediate elements (between args of function)
                self.result_string += next_function.separators[i]

                # add ith argument
                position = self.function_replacement(next_start_end)

                # if some argument is missing END_OF_FUNCTION
                if position == -1:
                    raise TranslationError(
                        f"Missing argument for latex function"
                        + f'\n        in "{self.latex_string}"'
                        + f'\n        at "{self.latex_string[next_function_start:]}"'
                        + "\n"
                    )

            # add last element of function
            self.result_string += next_function.separators[i + 1]

            position = self.function_replacement(position)
            return position

        raise Exception(f"Should not happen - Latex2LeanMath.function_replacement")

    def implicit_operations(self):
        still_matching = len(IMPLICIT_OPERATIONS)

        while still_matching > 0:
            still_matching = len(IMPLICIT_OPERATIONS)

            # match all patterns
            for pattern, operation in IMPLICIT_OPERATIONS.items():
                match_pattern = re.search(pattern, self.result_string)

                # if no match
                if match_pattern == None:
                    still_matching -= 1
                    continue

                # get the position for the added '*'
                end = match_pattern.end()
                right_term = match_pattern.groups()[-1]
                position = end - len(right_term)

                self.result_string = (
                    self.result_string[:position]
                    + operation
                    + self.result_string[position:]
                )

    def get_next_function(self, position: int) -> tuple[int, int, LatexFunction]:
        starts = []

        for tex_fun in LATEX_FUNCTIONS:
            match_function = re.search(tex_fun.pattern, self.latex_string[position:])

            # no match for this latex function
            if match_function == None:
                continue

            # add to list
            next_function_start = match_function.start() + position
            next_function_end = match_function.end() + position
            starts.append((next_function_start, next_function_end, tex_fun))

        if len(starts) == 0:
            return None, None, None

        return min(starts)

    def get_next_end(self, position: int) -> int:
        # check for next end
        match_end = re.search(END_OF_FUNCTION, self.latex_string[position:])

        # no end found
        if match_end == None:
            return None, None

        next_end_start = match_end.start() + position
        next_end_end = match_end.end() + position

        return next_end_start, next_end_end

    def get_next_start(self, position: int) -> int:
        # check for next start
        match_start = re.search(START_OF_FUNCTION, self.latex_string[position:])

        # no start found
        if match_start == None:
            return None, None

        next_start_start = match_start.start() + position
        next_start_end = match_start.end() + position

        return next_start_start, next_start_end

    def __str__(self) -> str:
        return self.result_string


def translate_latex_math(latex_string: str) -> str:
    return str(Latex2LeanMath(latex_string))



import textwrap

INDENTATION = "  "


def indent(text, level=1):
    return textwrap.indent(text, INDENTATION * level)


def nth(n):
    if n == 1:
        return "1st"
    if n == 2:
        return "2nd"
    if n == 3:
        return "3rd"
    return f"{n}th"


def subscript(n: int):
    subscripts = {
        "0": "₀",
        "1": "₁",
        "2": "₂",
        "3": "₃",
        "4": "₄",
        "5": "₅",
        "6": "₆",
        "7": "₇",
        "8": "₈",
        "9": "₉",
    }

    result = ""
    for digit in str(n):
        result += subscripts[digit]

    return result



from dataclasses import dataclass


@dataclass(frozen=True)
class LatexFunction:
    pattern: str
    separators: list[str]


@dataclass(frozen=True)
class MathSet:
    latex: str
    intermediate: str
    lean: str
    full_words: tuple[str]


"""
The characters used to group function arguments.
"""
START_OF_FUNCTION = "{"
END_OF_FUNCTION = "}"

"""
For each set, 
- the first argument should be its representation in LaTeX, 
- the second can be anything, an intermediate representation that is used. It should not be something that will be interpreted in another way in lean.
- the third argumetn is the representation in lean.
- the last argument contains a tuple of words that a user could type to refer to the set.
"""
SETS: dict[str, MathSet] = {
    "ℕ": MathSet(r"\mathbb{N}", "ℕ", "Nat", ("natural",)),
    # "ℤ": MathSet(r"\mathbb{Z}", "ℤ", "Int", ("integer",)), # not compatible with the current work, because header considers that we are working with Naturals
    # "ℚ": MathSet(r"\mathbb{Q}", "ℚ", "???", , ("rational,")), # Rationals not implemented in lean yet ?
    # "ℝ": MathSet(r"\mathbb{R}", "ℝ", "???", ("real,")), # Reals not implemented in lean yet ?
    # "ℂ": MathSet(r"\mathbb{C}", "ℂ", "???", ("complex,")), # Complexes not implemented in lean yet ?
}

"""
The keys will be replaced by the corresponding values when translating a math expression from LaTeX to Lean. 
"""
LATEX_SYMBOLS: dict[str, str] = {
    # remove new lines
    "\n": "",
    # operations
    r"\mod": r"%",
    r"\cdot": r"*",
    # parenthesis
    r"\bigg": r"",
    r"\left": r"",
    r"\right": r"",
    # equalities / inequalities
    r"\eq": r"=",
    r"\neq": r"≠",
    r"\leq": r"≤",
    r"\geq": r"≥",
    r"\lt": r"<",
    r"\gt": r">",
    # for sets, all the sets in SETS will be added at the end and should not be added here
    r"\in": r"∈",
    r"\times": r"×",
    # to allow for sets such as a, b \in \{a, b\}
    r"\{": r"{",
    r"\}": r"}",
}

"""
Here, the separators will be added before the first argument, between arguments and after the last argument. The pattern should be a regex pattern, you should therefore watch out for special characters. Some translations are specific to the translation system (e.g. [] will be recognized by ExpressionPossibilities).

The arguments of the LatexFunction dataclass are:
    - `pattern` : str, regex pattern to match the start of the LaTeX function.
    - `separators` : list[str], the separators before, between and after the arguments of the LaTeX function. There must therefore be one more separator than the number of arguments of the LaTeX function (e.g. the \\frac function has 2 arguments (numerator and denominator) and therefore 3 separators).
"""
LATEX_FUNCTIONS: list[LatexFunction] = [
    # watch out for special regex characters (^, $, ., ?, *, +, |, (, ), [, ], {, }, \)
    LatexFunction(r"\\frac *{", ["((", ") / (", "))"]),
    LatexFunction(r"\\sqrt *{", ["((", ") ^ (1/2))"]),
    LatexFunction(r"\^ *{", ["^(", ")"]),
    LatexFunction(r"_ *{", ["_", ""]),
    LatexFunction(r"∈ *{", ["∈ [", "]"]),
]

"""
Here, the values (implicit operation) will be placed before the key's last regex group. The keys should be a regex pattern, you should therefore watch out for special characters. There is no reason I can think of to have more than one group in the regex pattern, but the system will not fail if there are.
"""
IMPLICIT_OPERATIONS: dict[str, str] = {
    # 2(x + y) -> 2 * (x + y)
    r"\w *(\()": "*",
    # (x + y)2 -> (x + y) * 2
    r"\) *(\w)": "*",
    # 2x2 -> 2 * x2
    r"(?:^|\W)\d+ *( *[a-zA-Z]\w* *)": "*",
    # (x + y)(z + w) -> (x + y) * (z + w)
    r"\) *(\()": "*",
}


# ----------------------------------------------------------
# Adding sets to symbols
# ----------------------------------------------------------

for s in SETS.values():
    if s.latex in LATEX_SYMBOLS:
        raise ValueError(
            f"The translation {s.latex} => {s.intermediate} could not be added to the symbol dictionary because {s.latex} was already there."
        )

    LATEX_SYMBOLS[s.latex] = s.intermediate



class Translatable:
    def translate(self, **kwargs) -> str:
        """Returns the lean translation of the object."""
        raise NotImplementedError()

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        """Returns a list of tuples (`type`, `section`) describing the way the string was processed.

        The type can be one of the following:
        - `"keyword"`: the section of the string is a keyword used to interpret the string (e.g. "theorem", "have")
        - `"parameter"`: the section of the string is a parameter used in the translation (e.g. a theorem name, a math expression)
        - `"ignored"`: the section of the string was ignored (most filling words like "we", "therefore", ...)
        """
        raise NotImplementedError()

    def can_create_new_goals(self) -> bool:
        """Returns whether the object should be allowed to create new goals."""
        return False

    def change_status(self) -> bool:
        """Returns whether the object should change the status (e.g. if doing an induction on a variable)."""
        return False



import re
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Callable, Union
from dataclasses import dataclass
from ..utils.exceptions import LeanError
from ..utils.text import indent, nth
from ..proof_elements.statement.statement import Statement
from ..proof_elements.statement.induction import Induction
from ..proof_elements.theorem.theorem import Theorem


@dataclass(frozen=True)
class LeanBlock:
    """A block of results from lean, each one containing variables, hypotheses and a goal."""

    # (name, set), variables can contain multiple identifiers
    variables: list[tuple[str, str]]
    # (hypothesis name, expression), each hypothesis can only contain one identifier
    hypotheses: list[tuple[str, str]]
    # the goal for this block
    goal: str

    def __str__(self) -> str:
        # for missing cases
        if not self.variables and not self.hypotheses:
            return f"{self.goal} (missing case)"

        # variables
        if self.variables:
            variables = "\n".join(f"{id_} : {set_}" for id_, set_ in self.variables)
            var_block = f"Variables we can use :\n{indent(variables)}\n"
        else:
            var_block = ""

        # hypotheses
        if self.hypotheses:
            hypotheses = "\n".join(
                f"{name_} : {expr_}" for name_, expr_ in self.hypotheses
            )
            hyp_block = (
                f"What we have shown so far (for this goal) :\n{indent(hypotheses)}\n"
            )
        else:
            hyp_block = ""

        return self.goal + "\n" + var_block + hyp_block


@dataclass
class State:
    goals: list[LeanBlock]
    last_statement: Union[Theorem, Statement]
    lean_text: str

    def __str__(self):
        if len(self.goals) == 0:
            return ""

        if len(self.goals) == 1:
            return f"To finish this proof, we need to prove this: {indent(str(self.goals[0]))}"
        
        result = f"To finish this proof, we need to solve these {len(self.goals)} cases (goals):"
        for i, goal in enumerate(self.goals):
            result += f"\n{nth(i+1)} case: solve {indent(str(goal))}"

        return result


@dataclass
class SpecificError:
    pattern: str
    statement_check: Callable
    added_blocks: Callable


# if any of ERRORS is matched, the result will be FAIL, and the system will cancel the last input.
ERRORS = [
    r"error: tactic .+ failed",
    r"error: ring failed",
    r"error: unknown .+",
    r"error: expected .+",
    r"error: invalid .+",
    r"error: .+ is missing",
    r"error: type mismatch",
]

SPECIFIC_ERRORS = [
    SpecificError(
        pattern=r"error: unexpected end of input;",
        statement_check=lambda st: isinstance(st, Induction),
        added_blocks=lambda st: [
            LeanBlock(
                variables=[], hypotheses=[], goal=f"values of {st.variable.translate()}"
            )
        ],
    )
]

# patterns need a fullmatch on a line to work
# patterns to recognize a goal, this will be checked first, and will create a new LeanBlock at each match. The only group must contain the goal.
GOAL_PATTERNS = [r"⊢ (.*)"]
# patterns to recognize variables, these will be checked second. The first group must contain the identifier(s) and the second must contain the corresponding set.
VAR_PATTERNS = [
    r"(.*) : ([ℕℤℚℝ].*)",
]
# patterns to recognize hypotheses, these will be checked last. the first group contains hypothesis name and the second contains the set.
HYP_PATTERNS = [r"(.*) : (.*)"]

# missing cases are all lines between "missing cases" and "././" or "error" (other error)
MISSING_CASES_START = r"missing cases:\n"
MISSING_CASES_STOP = r"(?:\./\./|error:)"

MAIN_FILE = "Main.lean"
BUILD_COMMAND = "lake build"


def lean_feedback(
    input: str, statement: Statement, project_directory: Path
) -> list[LeanBlock]:
    """Gets the feedback from lean.

    Args:
        input (str): the lean text to be checked

    Returns:
        list[LeanBlock]: a list of lean blocks, each containing variables, hypotheses and goals.
    """

    # get feedback
    feedback = raw_feedback(input, project_directory)

    # errors
    match = match_list(ERRORS, feedback, type="search")
    if match is not None:
        raise LeanError(f"Lean error: {feedback[match.start():match.end()]}\n")

    # separate blocks
    lean_blocks = separate_elements(feedback)

    # check missing cases
    missing_cases = get_missing_cases(feedback)

    # check specific errors
    specific_errors = check_specific_errors(statement, feedback)

    return specific_errors + lean_blocks + missing_cases


def raw_feedback(input: str, project_directory: Path) -> str:
    """Gets the raw feedback from lean.

    Args:
        input (str): the text to be fed to lean.

    Returns:
        str: the raw (terminal) feedback
    """

    with open(project_directory / MAIN_FILE, "w") as f:
        f.write(input)

    res = Popen(
        BUILD_COMMAND,
        shell=True,
        cwd=project_directory,
        stdout=PIPE,
        stderr=PIPE,
    )

    return res.stdout.read().decode("utf-8") + res.stderr.read().decode("utf-8")


# -------------------------------------------------------------------------
# HELPER FUNCTIONS
# -------------------------------------------------------------------------


def match_list(patterns: list[str], text: str, type: str = "search") -> re.Match:
    """Returns the first match of a list of patterns.

    Args:
        patterns (list[str]): a list of patterns to try
        text (str): the text to match
        type (str, optional): the type of call, any of "search", "fullmatch", "match". Defaults to "search".

    Returns:
        bool: _description_
    """
    if type not in ["search", "fullmatch", "match"]:
        raise Exception(f"type must be one of 'search', 'fullmatch', 'match'")
    if type == "search":
        call = re.search
    if type == "fullmatch":
        call = re.fullmatch
    if type == "match":
        call = re.match

    # check each pattern
    for pattern in patterns:
        match = call(pattern, text)
        # return first match
        if match is not None:
            return match

    # no match found
    return None


def separate_elements(raw_feedback: str) -> list[LeanBlock]:
    """Separates the different elements, based on `GOAL_PATTERNS`, `SECONDARY_GOAL_PATTERNS`, `VAR_PATTERNS` and `HYP_PATTERNS`.

    Args:
        raw_feedback (str): the feedback from lean.

    Returns:
        list[LeanBlock]: a list of goal structures.
    """

    # initialize variables
    blocks: list[LeanBlock] = []
    current_variables: list[tuple[str, str]] = []
    current_hypotheses: list[tuple[str, str]] = []

    # iterate over each line
    for line in raw_feedback.splitlines():
        # goal
        match = match_list(GOAL_PATTERNS, line, type="fullmatch")
        if match is not None:
            # create a new block
            blocks.append(
                LeanBlock(
                    variables=current_variables,
                    hypotheses=current_hypotheses,
                    goal=match.group(1),
                )
            )
            # clean the current variables and hypotheses
            current_variables = []
            current_hypotheses = []
            # skip to next line
            continue

        # variable
        match = match_list(VAR_PATTERNS, line, type="fullmatch")
        if match is not None:
            # add the variable to the current variables
            current_variables.append((match.group(1), match.group(2)))
            # skip to next line
            continue

        # hypothesis
        match = match_list(HYP_PATTERNS, line, type="fullmatch")
        if match is not None:
            # add each hypothesis separately
            for hyp in match.group(1).split():
                current_hypotheses.append((hyp, match.group(2)))
            # skip to next line
            continue

    return blocks


def get_missing_cases(raw_feedback: str) -> list[LeanBlock]:
    missing_cases = []

    match = re.search(MISSING_CASES_START, raw_feedback)
    if not match:
        return []

    for line in raw_feedback[match.end() :].splitlines():
        if re.match(MISSING_CASES_STOP, line):
            break

        missing_cases.append(LeanBlock([], [], line))

    return missing_cases


def check_specific_errors(statement: Statement, raw_feedback: str) -> list[LeanBlock]:
    if not statement:
        return []

    specific_blocks = []
    for error in SPECIFIC_ERRORS:
        if re.search(error.pattern, raw_feedback) and error.statement_check(statement):
            specific_blocks += error.added_blocks(statement)

    return specific_blocks



import re
from .algebra import Algebra
from ..utils.exceptions import MatchingError


class Expression(Algebra):
    # anything (an expression can be only a single term, a numeric value or a computation (x+y))
    pattern: str = r".*"

    def __init__(self, string: str) -> None:
        # possible errors
        if string.strip() == "":
            raise MatchingError(
                f"Could not match empty string in {self.__class__.__name__}"
            )
        if re.match(r"[≠=<≤>≥]", string):
            raise MatchingError(
                f"Could {self.__class__.__name__} cannot contain equality operators"
            )

        self.string = string.strip()

    def translate(self, **kwargs) -> str:
        return self.string



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="natural2lean",
    version="0.0.3",
    author="Augustin d'Oultremont",
    author_email="augustin.doultremont@outlook.com",
    description="Translation of proofs from Natural Language to Lean",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Augustindou/natural2lean",
    project_urls={
        "Bug Tracker": "https://github.com/Augustindou/natural2lean/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src", exclude=["*-old"]),
    python_requires=">=3.9",
    install_requires=[],
    extras_require={
        "dev": [
            "pytest>=3.6",
        ],
    },
)



import re
from .translate_math import translate_latex_math
from ..utils.exceptions import MatchingError
from ..utils.translatable import Translatable


space = r"\s*"


class Algebra(Translatable):
    pattern: str = space.join(
        [
            "",  # ignore leading space
            r"\${1,2}",
            r"(.+?)",
            r"\${1,2}",
            "",  # ignore trailing space
        ]
    )

    def __init__(self, string: str, match_type: str = "full"):
        if match_type.lower() in ("fullmatch", "full"):
            matcher = re.fullmatch
        elif match_type.lower() in ("search", "partial"):
            matcher = re.search
        else:
            raise ValueError(f"match_type should be either 'full' or 'search'")
        if not (match := matcher(self.pattern, string)):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        self.latex_string = f"$ {match.group(1).strip()} $"
        self.lean_string = translate_latex_math(self.latex_string.strip("$"))



import re
from typing import Callable
from ..utils.exceptions import MatchingError
from ..algebra.algebra import Algebra
from ..algebra.identifiers import MultipleIdentifiers

# TODO COMMENTS
REPLACEMENTS: dict[str, str] = {
    r"\$\$": "$",
    r"\$\s*and\s*\$": ", ",
    r"\$\s*,\s*\$": ", ",
}
SEPARATORS: list[str] = [",", "and"]
NEGATIONS: list[str] = ["not", "n't"]

# tuple of function name, function pattern, and callable to rearrange function arguments (keep_order and invert_order are examples, but one can also directly use a lambda function for more specific needs)
# pattern must contain one group for each argument
keep_order = lambda *args: args
invert_order = lambda *args: args[::-1]
# even groups are arguments, all groups will be highlighted by interpretation_feedback, use empty groups if needed ;)
FUNCTIONS: list[tuple[str, str, Callable]] = [
    ("even", r"(\$.+?\$).*?(even)", keep_order),
    ("odd", r"(\$.+?\$).*?(odd)", keep_order),
    (
        "divisible",
        r"(\$.+?\$).*?(divisible\s*by).*?(\$.+?\$)",
        invert_order,
    ),
]
# ------------- VALIDITY CHECKS -------------


def matches_algebra(string: str, cls: type) -> bool:
    try:
        cls(get_lean_math(string))
        return True
    except MatchingError:
        return False


VALIDITY_CHECKS: list[Callable] = [
    # even number of $ signs (avoid splitting inside a math mode)
    lambda s: s.count("$") % 2 == 0,
    # no identifiers alone
    lambda s: not matches_algebra(s, MultipleIdentifiers),
]


# ---------------- SMALL FUNCTIONS -----------------
def get_lean_math(string: str) -> str:
    al = Algebra(apply_replacements(string))
    return al.lean_string


def apply_replacements(string: str) -> str:
    """Groups identifiers ($a$ and $b$) into a single math mode ($a, b$).

    Args:
        string (str): the input string
    """
    for pattern, replacement in REPLACEMENTS.items():
        while (match := re.search(pattern, string)) != None:
            string = string[: match.start()] + replacement + string[match.end() :]
    return string



from .induction import Induction
from .induction_case import InductionCase
from .simple_statements import SimpleStatement
from .such_that import SuchThat
from .have import Have
from .proof_structure import ProofStructure
from .statement import Statement
from ...propositions.multiple_propositions import MultiplePropositions
from ...utils.exceptions import MatchingError, TranslationError

POSSIBILITIES = [InductionCase, Induction, Have, ProofStructure]
# conclusions, statements that would only be accepted if they are conclusions of a goal
CCL_POSSIBILITIES = [MultiplePropositions, SuchThat, SimpleStatement]


def get_statement(string: str, **kwargs) -> Statement:
    for poss in POSSIBILITIES + CCL_POSSIBILITIES:
        try:
            return poss(string, **kwargs)
        except MatchingError:
            pass

    raise TranslationError(
        f"Could not match any statement for '{string}', tried "
        + ", ".join([p.__name__ for p in POSSIBILITIES])
    )



import re
from .theorem import Theorem, STATEMENT_POSSIBILITIES
from ...utils.exceptions import MatchingError

space = r"\s*"


class NamedTheorem(Theorem):
    pattern: str = space.join(
        [
            "",  # ignore leading space
            r"([Tt]heorem|[Ll]emma)",  # Theorem keyword
            r"(.*)",  # Theorem name
            r"(:)",  # separation between name and statement
            r"(.*)",  # Theorem statement
            "",  # ignore trailing space
        ]
    )

    def __init__(self, string: str, **kwargs) -> None:
        if not (match := re.fullmatch(self.pattern, string)):
            raise MatchingError(
                f"Could not match {string} in {self.__class__.__name__}"
            )

        # keep it for processing in interpretation_feedback()
        self.match = match

        # name
        self.latex_name = match.group(2)
        self.lean_name = self.latex_name.lower().replace(" ", "_")
        if self.lean_name[0].isdigit():
            self.lean_name = "_" + self.lean_name

        # content
        for poss in STATEMENT_POSSIBILITIES:
            try:
                self.statement = poss(match.group(4))
                self.set_hypotheses_and_theses()
                return
            except MatchingError:
                pass

        raise MatchingError(
            f"Could not match any statement to prove for {string}, tried {', '.join([p.__name__ for p in STATEMENT_POSSIBILITIES])}."
        )

    def interpretation_feedback(self) -> list[tuple[str, str]]:
        feedback = [
            ("keyword", self.match.group(1)),
            ("parameter", self.match.group(2)),
            ("keyword", self.match.group(3)),
        ]

        return feedback + self.statement.interpretation_feedback()

    # translate method is inherited
